; -----------------------------------------------------------------------------
; BIFROST* ENGINE by Einar Saukas - v1.0
; A Rainbow Graphics Support Engine for Animated Tiles
;
; Based on the ZXodus Engine by Andrew Owen
; Most 16x16 tiles created by Dave Hughes (R-Tape)
;
; To be compiled with PASMO - http://pasmo.speccy.org/
; -----------------------------------------------------------------------------

; Animation speed: 2 or 4 frames per second
ANIM_SPEED     EQU 4

; Animation size: 2 or 4 frames per animation group
ANIM_GROUP     EQU 4

; First non-animated frame
STATIC_MIN     EQU 128

; Value subtracted from non-animated frames
STATIC_OVERLAP EQU 128

; Location of the tiles table (64 bytes per tile)
TILE_IMAGES    EQU 48500

; Location of the tile map (9x9=81 tiles)
TILE_MAP      EQU 65283

; -----------------------------------------------------------------------------
MACRO LD_DE_INDEXED, addr, n, delta
base DEFL addr
REPT n
.SHIFT
ENDM
        ld      de, base+delta
ENDM

; -----------------------------------------------------------------------------
        org     56768
main_engine:
; preserve all registers
        push    af
        push    bc
        push    de
        push    hl
        ex      af, af'
        exx
        push    af
        push    bc
        push    de
        push    hl
; render and animate 3 tiles
        ld      a, $c6
        ld      (animate_tile+2), a
        call    draw_tiles
; render (and perhaps animate) another 3 tiles
IF ANIM_SPEED=4
        ld      a, $c6
ELSE
        ld      a, $fe
ENDIF
        ld      (animate_tile+2), a
        call    draw_tiles+(get_tile-draw_tiles)/27
; synchronize with the raster beam
        ld      bc, $070b
sync_raster:
        dec     c
        jr      nz, sync_raster
        ld      c, 12
        ld      a, ($4000)
        djnz    sync_raster
; wait for the raster beam
        ld      b, 5
wait_raster:
        djnz    wait_raster
; preserve stack pointer
        ld      (sp_store), sp
; race the raster beam to update attributes at the right time
race_raster:
REPT 18, ROWREPT
REPT 8
        ld      sp, $5833+(32*ROWREPT)
        ld      bc, 0                   ; columns 01 and 02
        ld      de, 0                   ; columns 03 and 04
        ld      hl, 0                   ; columns 05 and 06
        exx
        ld      de, 0                   ; columns 07 and 08
        ld      hl, 0                   ; columns 09 and 10
        ld      bc, 0                   ; columns 17 and 18
        push    bc
        ld      bc, 0                   ; columns 15 and 16
        push    bc
        ld      bc, 0                   ; columns 13 and 14
        push    bc
        ld      bc, 0                   ; columns 11 and 12
        push    bc
        push    hl
        push    de
        exx
        push    hl
        push    de
        push    bc
ENDM
ENDM
; restore stack pointer
        ld      sp, (sp_store)
; restore all registers
        pop     hl
        pop     de
        pop     bc
        pop     af
        exx
        ex      af, af'
        pop     hl
        pop     de
        pop     bc
        pop     af
        jp      $38

; -----------------------------------------------------------------------------
draw_tiles:
REPT 27, CALLREPT
REPT 3, DRAWREPT
tilenum DEFL (CALLREPT*3+DRAWREPT)*7%81
rownum  DEFL tilenum/9
colnum  DEFL tilenum%9
        ld      hl, TILE_MAP + tilenum
        call    get_tile
        ld      de, $4001 + (rownum*2+1)/8*2048 + (rownum*2+1)%8*32 + colnum*2
        call    draw_bmp
        ld      de, $4001 + (rownum*2+2)/8*2048 + (rownum*2+2)%8*32 + colnum*2
        call    draw_bmp
        LD_DE_INDEXED   race_raster + rownum*2*41*8, colnum, 4,7,10,14,17,32,28,24,20
        call    draw_attr
ENDM
        pop     hl
        push    hl
        dec     hl
        ld      (hl), (draw_tiles + (CALLREPT+2)%27*(get_tile-draw_tiles)/27)/256
        dec     hl
        ld      (hl), (draw_tiles + (CALLREPT+2)%27*(get_tile-draw_tiles)/27)%256
        ret
ENDM

; -----------------------------------------------------------------------------
get_tile:
        ld      a,(hl)
        cp      STATIC_MIN
        jp      c, animate_tile
        inc     a
        jr      z, skip_tile
        sub     1+STATIC_OVERLAP
        jr      lookup_tile
animate_tile:
        rrca
IF ANIM_GROUP=4
        rrca
        add     a, $40
        rlca
ELSE
        nop
        add     a, $80
        nop
ENDIF
        rlca
        ld      (hl), a
lookup_tile:
        ld      l, a
        ld      h, 0
        add     hl, hl
        add     hl, hl
        add     hl, hl
        add     hl, hl
        add     hl, hl
        add     hl, hl
        ld      de, TILE_IMAGES
        add     hl, de
        ret
skip_tile:
        ld      b, 170
delay_tile:
        djnz    delay_tile
        ld      a, i                    ; extra delay
; skip drawing
        pop     hl
        ld      de, 18
        add     hl, de
        jp      (hl)

; -----------------------------------------------------------------------------
draw_bmp:
        ld      bc, $0410
next_bmp:
REPT 2
        ldi
        ldi
        dec     e
        dec     e
        inc     d
ENDM
        djnz    next_bmp
        ret

; -----------------------------------------------------------------------------
draw_attr:
        ld      bc, $1020
next_attr:
        ldi
        ldi
        ld      a, e
        add     a, 39
        ld      e, a
        ld      a, d
        adc     a, 0
        ld      d, a
        djnz    next_attr
        ret

; -----------------------------------------------------------------------------
        org $fdef
; RAND USR 65007 to activate engine
        di
        ld      a, $fe
        ld      i, a
        im      2
        ei
        ret

; -----------------------------------------------------------------------------
; RAND USR 65016 to deactivate engine
        di
        im      1
        ei
        ret

; -----------------------------------------------------------------------------
; interrupt address at $fdfd
        jp      main_engine

; -----------------------------------------------------------------------------
; jump vector table at addresses $fe00-$ff00
REPT 257
        defb    $fd
ENDM

; -----------------------------------------------------------------------------
; preserve cache address
sp_store:
        defw    0

; -----------------------------------------------------------------------------
