; ZXodus II Engine (preview)
; Copyright (c) 2012 Andrew Owen

; Uses local variables. Compile with: pasmo --alocal zx2.asm zx2.bin

; RANDOMIZE USR 24576

; ------------------------------------------------------------------------------
; INCORPORATING ADAPTATIONS OF
; Optimization of the original ZXodus Engine by Simon Owen.
; Further optimization by Einar Saukas.
; Channel Creator by Ian Beardsmore (Your Spectrum #7, September 1984).
; Proportional Print by Tony Samuels (Your Spectrum #13, April 1985).
; Interrupt Manager by Toni Baker (Your Spectrum #17, August 1985).
; PT3 Player by Sergey Bulba.
; The Ultima IV tileset by Richard Garriot.
; The Ultima IV (MSX) font by Newtopia Planning.
; Stones by David R. Watson & Kathleen Jones.
; Shiru's BeepFX Engine.
; Marko's Force USR0 code.

; ------------------------------------------------------------------------------
; MEMORY MAP
; Correct contention (+3/+2B) is assumed (n' = contended).

; 0 - tiles (0-255)
; 1 - world map (128*128)
; 2 - main code (fast RAM)
; 3 - town maps 16 * (32*32)
; 4'- not allocated
; 5'- screen, initialization code, tables, music buffer
; 6'- music (compressed)
; 7'- auxilliary code (slow RAM)

; Note: The page stored in port_7ffd will be set when the IM2 routine returns
; control.

; ------------------------------------------------------------------------------
; CONSTANTS
; You can change these to adapt the engine to your own needs.

tile_gfx	equ	49152		; now stored in bank 0
line_size	equ	41			; 41 bytes per rainbow line
grid_offset	equ	1			; display offset for left edge
num_rows	equ	9
num_cols	equ	9
tile_height	equ	16
num_tiles	equ	num_rows*num_cols
row_offset	equ	line_size*tile_height

report_j	equ	0x15c4
chan_open	equ	0x1601
make_room	equ	0x1655
pixel_add	equ	0x22aa
strms		equ	0x5c10
chans		equ	0x5c4f
prog		equ	0x5c53
df_cc		equ	0x5c84
s_posn		equ	0x5c88
attr_p		equ	0x5c8d

strm_n		equ	4						; stream #4
strm_offst	equ	0x5c10+((strm_n+3)*2)	; address of channel offset	in streams
										; table

; ------------------------------------------------------------------------------
; BANK 5

	org		0x4000
	defs	0x4000

; ------------------------------------------------------------------------------
; SETUP
; Sets up, IM2 code, 6x8 font channel and stream (#4), background

	org		0x6000
	
initi:
	di
	push	af				; save registers
	push	bc
	push	de
	push	hl
	exx
	push	af
	push	bc
	push	de
	push	hl

c_chan:
	ld		hl, (prog)		; store system variable PROG in HL
	dec		hl				; move hl down one address
	ld		bc, 5			; allocate 5 bytes for channel below BASIC area
	push	bc
	call	make_room		; call the MAKE-ROOM routine
	pop		bc
	ld		hl, ch_data + 4
	lddr					; copy ch_data to new channel space
	ld		hl, (chans)		; store system variable CHANS in HL
	ex		de, hl
	inc		hl
	inc		hl				; now HL = allocated address + 1
	sbc		hl, de			; calculate offset between start of channels
							; area and start of the new channel space
							; (notice the carry flag was already cleared
							; from executing CALL make_room earlier)
	ld		(strm_offst), hl; attach stream by storing channel address
							; offset in streams table

	xor		a				; set BORDER 0
	out		(0xfe), a
	ld		hl, border		; setup ingame screen
	ld		de, 0x4000
	call	depack

	ld		a, 0x80			; Loads the A registers with the high byte of the
							; interrupt vector (INTVC).
	ld		i, a			; Copies the value in the A register into the
							; interrupt register.
	im		2				; Selects interrupt mode 2 - this is the mode that
							; allows you to use your own interrupt routines on
							; the Spectrum ratehr than the one in the ROM.

	ld		bc, 0x7ffd		; Paging.
	ld		a, 22			; bank 6 (music)
	out		(c), a

	ld		hl, 0xc000		; depack tune
	ld		de, music
	call	depack

	call	p3x_init		; Initialize music.

	pop		hl				; restore registers
	pop		de
	pop		bc
	pop		af
	exx
	pop		hl
	pop		de
	pop		bc
	pop		af
	ei

null:
	ret

ch_data:
	defw	print6x8		; address of the PRINT # routine
	defw	report_j		; address of the INPUT # routine
	defb	'S'				; channel type 'S'

; ------------------------------------------------------------------------------
; BORDER
; The screen border (packed to 561 bytes)

border:
	incbin	"start.pck"

; ------------------------------------------------------------------------------
; FONT (832 bytes)
; The 6x8 typeface

font:
	incbin	"charset.6x8"

; IMPORTANT: variables atflg, coords_x, coords_y must be defined consecutively
;            in this order!

atflg:
	defb	0

coords_x:
	defb	160
	
coords_y:
	defb 	0

chrad:
	defw	0


; ------------------------------------------------------------------------------
; ROW table (192 bytes)
; offsets for scrolling the main text window

row_table:
	defw	0x4894
	defw	0x4994
	defw	0x4a94
	defw	0x4b94
	defw	0x4c94
	defw	0x4d94
	defw	0x4e94
	defw	0x4f94
	defw	0x48b4
	defw	0x49b4
	defw	0x4ab4
	defw	0x4bb4
	defw	0x4cb4
	defw	0x4db4
	defw	0x4eb4
	defw	0x4fb4
	defw	0x48d4
	defw	0x49d4
	defw	0x4ad4
	defw	0x4bd4
	defw	0x4cd4
	defw	0x4dd4
	defw	0x4ed4
	defw	0x4fd4
	defw	0x48f4
	defw	0x49f4
	defw	0x4af4
	defw	0x4bf4
	defw	0x4cf4
	defw	0x4df4
	defw	0x4ef4
	defw	0x4ff4
	defw	0x5014
	defw	0x5114
	defw	0x5214
	defw	0x5314
	defw	0x5414
	defw	0x5514
	defw	0x5614
	defw	0x5714
	defw	0x5034
	defw	0x5134
	defw	0x5234
	defw	0x5334
	defw	0x5434
	defw	0x5534
	defw	0x5634
	defw	0x5734
	defw	0x5054
	defw	0x5154
	defw	0x5254
	defw	0x5354
	defw	0x5454
	defw	0x5554
	defw	0x5654
	defw	0x5754
	defw	0x5074
	defw	0x5174
	defw	0x5274
	defw	0x5374
	defw	0x5474
	defw	0x5574
	defw	0x5674
	defw	0x5774
	defw	0x5094
	defw	0x5194
	defw	0x5294
	defw	0x5394
	defw	0x5494
	defw	0x5594
	defw	0x5694
	defw	0x5794
	defw	0x50b4
	defw	0x51b4
	defw	0x52b4
	defw	0x53b4
	defw	0x54b4
	defw	0x55b4
	defw	0x56b4
	defw	0x57b4
	defw	0x50d4
	defw	0x51d4
	defw	0x52d4
	defw	0x53d4
	defw	0x54d4
	defw	0x55d4
	defw	0x56d4
	defw	0x57d4
	defw	0x50f4
	defw	0x51f4
	defw	0x52f4
	defw	0x53f4
	defw	0x54f4
	defw	0x55f4
	defw	0x56f4
	defw	0x57f4
	defw	clear_row
	defw	clear_row
	defw	clear_row
	defw	clear_row
	defw	clear_row
	defw	clear_row
	defw	clear_row
	defw	clear_row

clear_row:
	defb	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0

; ------------------------------------------------------------------------------
; MUSIC BUFFER

	org		30720

music:
	defs	2048				; buffer must be as long as the longest tune

; ------------------------------------------------------------------------------
; BANK 2

; ------------------------------------------------------------------------------
; INTERRUPT VECTORS
; These 257 bytes store the address of the Manager program.

	org		0x8000

intvc:
	defs	257, 0x81		; Can't trust devices not to put a byte on the bus	
							; and 0xffff could be paged out during an interrupt.

; ------------------------------------------------------------------------------
; DEBITBUST
; Call with source in HL, destination in DE

	org		0x8101

depack:
	ld	a,128
	
	exx
	ld	de,1
	exx
	
depack_loop:
	call getbits
	jr	c,output_compressed	;if set, we got lz77 compression
	ldi				;copy byte from compressed data to destination (literal byte)
	jr	depack_loop
	

;handle compressed data
output_compressed:
	ld	c,(hl)		;get lowest 7 bits of offset, plus offset extension bit
	inc	hl		;to next byte in compressed data

output_match:
	ld	b,0
	bit	7,c
	jr	z,output_match1	;no need to get extra bits if carry not set

	call getbits
	call rlbgetbits
	call rlbgetbits
	call rlbgetbits

	jr	c,output_match1	;since extension mark already makes bit 7 set 
	res	7,c		;only clear it if the bit should be cleared
output_match1:
	inc	bc
	
	
;return a gamma-encoded value
;length returned in HL
	exx			;to second register set!
	ld	h,d
	ld	l,e             ;initial length to 1
	ld	b,e		;bitcount to 1

;determine number of bits used to encode value
get_gamma_value_size:
	exx
	call getbits
	exx
	jr	nc,get_gamma_value_size_end	;if bit not set, bitlength of remaining is known
	inc	b				;increase bitcount
	jr	get_gamma_value_size		;repeat...

get_gamma_value_bits:
	exx
	call getbits
	exx
	
	adc	hl,hl				;insert new bit in HL
get_gamma_value_size_end:
	djnz	get_gamma_value_bits		;repeat if more bits to go

get_gamma_value_end:
	inc	hl		;length was stored as length-2 so correct this
	exx			;back to normal register set
	
	ret	c
;HL' = length

	push	hl		;address compressed data on stack

	exx
	push	hl		;match length on stack
	exx

	ld	h,d
	ld	l,e		;destination address in HL...
	sbc	hl,bc		;calculate source address

	pop	bc		;match length from stack

	ldir			;transfer data

	pop	hl		;address compressed data back from stack

	jr	depack_loop

rlbgetbits
	rl b
getbits
	add	a,a
	ret	nz
	ld	a,(hl)
	inc	hl
	rla
	ret

; spare bytes

; ------------------------------------------------------------------------------
; INTERRUPT MANAGER
; The Interupt Manager program starts here. This is called by the interrupt
; procedure fifty times a second.

	org		0x8181

mnger:
	push 	af					; save registers
	push 	bc
	push 	de
	push 	hl
	ex		af, af'				;'
	exx
	push 	af
	push 	bc
	push 	de
	push 	hl

	ld		hl, im_tab		; Interrupt manager table.
	ld		b, 0x10			; Loop counter (16 entries).

mnlop:
	ld		e, (hl)			; Loads DE with the routine address (if specified).
	inc		l				; Edge aligned table -- ignoring H saves 2 t-states
	ld		d, (hl)
	inc		l
	ld		c, (hl)			; Loads C with the flags for the routine.
	inc		l				; Point to COUNT.
	sra		c
	jr		nc, nonow		; Jumps if the routine is de-activated.
	dec		(hl)			; Decrements COUNT.
	jr		nz, nonow		; Jumps unless the routine has priority.
	inc		l				; Point to PRIORITY.
	ld		a, (hl)			; Loads A with the normal routine priority.
	dec		l				; Point to COUNT.
	ld		(hl), a			; Stores the priority in COUNT.
	push	bc				; Store the loop counter.
	push	hl				; Store the pointer to COUNT.
	ex		de, hl			; Places the routine address into HL.
	call	mncal			; Calls the routine.
	pop		hl				; Restore the pointer to COUNT.
	pop		bc				; Restore the loop counter.

nonow:
	inc		l				; Advance to next table entry.
	inc		l
	djnz	mnlop			; Back around to deal with next entry.

	ld		bc, 0x7ffd		; Paging.
	ld		a, 17			; Restore last value written (default to bank 1).
port_7ffd equ $-1
	out		(c), a
	pop		hl
	pop		de
	pop		bc
	pop		af
	ex		af, af'			;'
	exx
	pop		hl
	pop		de
	pop		bc
	pop		af
	ei						; Enable interrupts.
	ret

; ------------------------------------------------------------------------------
; ROUTINE VECTOR TABLE
; 21 three byte entries

;	33217 / 0x81c1
	jp		null

;	33220 / 0x81c4
	jp		null

;	33223 / 0x81c7
	jp		null

;	33226 / 0x81ca
	jp		null

;	33229 / 0x81cd
	jp		null

;	33232 / 0x81d0
	jp		null

;	33235 / 0x81d3
	jp		null

;	33238 / 0x81d6
	jp		null

;	33241 / 0x81d9
	jp		null

;	33244 / 0x81dc
	jp		null

;	33247 / 0x81df
	jp		null

;	33250 / 0x81e2
	jp		null

;	33253 / 0x81e5
	jp		null

;	33256 / 0x81e8
	jp		null

;	33259 / 0x81eb
	jp		null

;	33262 / 0x81ee
	jp		null

;	33265 / 0x81f1
	jp		null

;	33268 / 0x81f4
	jp		null

;	33271 / 0x81f7
	jp		null

;	33274 / 0x81fa
	jp		null

;	33277 / 0x81fd
	jp		null


; ------------------------------------------------------------------------------
; TILE MAP
; 81 byte buffer

	org		0x8200

tile_map:
	defb	8,8,8,4,4,4,4,4,7
	defb	8,9,4,36,4,13,14,15,4
	defb	8,4,40,248,33,4,4,4,4
	defb	4,4,4,47,4,4,2,2,4
	defb	11,2,2,4,4,2,1,1,2
	defb	2,1,1,2,2,1,16,1,2
	defb	2,1,1,1,1,1,1,2,3
	defb	10,2,1,1,1,1,2,3,5
	defb	4,4,2,130,1,2,3,5,6

; ------------------------------------------------------------------------------
; ROW / COLUMN TABLE
; 162 bytes for tile to row(*2) and column look-up table

rowcol_tab:
REPT num_tiles, tile
	defb (tile/num_cols)*2, tile%num_cols	 ; row*2, col
ENDM

; ------------------------------------------------------------------------------
; COLUMN TO RAINBOW OFFSET TABLE
; 9 bytes for column to rainbow code offset look-up table

offset_tab:
	defb	4, 7, 10, 26, 29, 32, 14, 17, 20

; ------------------------------------------------------------------------------
; MISC VARIABLES
; and a psuedo instruction

tile_num:
	defb	0				; next tile to draw

skip_num:
	defb	0

setup:
	defb	0				; used by ProTracker 3 Extreme II

mncal:
	jp		(hl)			; call to perform a pseudo CALL (HL)

; ------------------------------------------------------------------------------
; ROW TO BASE ADDRESS LOOK-UP TABLE
; 54 bytes for row to base addresses (attr, data2, data1) look-up table

	org		0x8300

addr_tab:
	defw rainbow+row_offset*0, 0x4041, 0x4021
	defw rainbow+row_offset*1, 0x4081, 0x4061
	defw rainbow+row_offset*2, 0x40c1, 0x40a1
	defw rainbow+row_offset*3, 0x4801, 0x40e1
	defw rainbow+row_offset*4, 0x4841, 0x4821
	defw rainbow+row_offset*5, 0x4881, 0x4861
	defw rainbow+row_offset*6, 0x48c1, 0x48a1
	defw rainbow+row_offset*7, 0x5001, 0x48e1
	defw rainbow+row_offset*8, 0x5041, 0x5021

; ------------------------------------------------------------------------------
; INTERRUPT MANAGER TABLE
; 16 interrupts (0-15) each have 5 bytes of data stored as follows:
;	0-1:	address			; in little endian format
; 	  2:	flags			; 0 = deactivated, 1 = activated
;	  3:	count			; must be initialized to the same value as priority
;	  4;	priority		; execute every 1 to 256 frames (0x01 - 0x00)  

im_tab:

	defw	rainbow_proc	; Rainbow processor.
	defb	1, 1, 1			; activated, trigger every interrupt.
							; 33592 - on / off

	defw	play			; AY player.
	defb	1, 2, 1			; activated, trigger every interrupt.
							; 33597 - on / off

	defw	scroll_tiles	; Scroll water and magic fields.
	defb	1, 3, 18		; Activated, trigger every 18th interrupt.
							; 33602 - on / off

	defw	animate_flags	; Wave flags on ships and buildings.
	defb	1, 4, 18		; Activated, trigger every 18th interrupt.
							; 33607 - on / off

	defw	animate_campfire; Animate the fire.
	defb	1, 5, 18		; Activated, trigger every 18th interrupt.
							; 33612 - on / off

	defw	swap_tiles		; Animate sprites.
	defb	1, 6, 18		; Activated, trigger every 18th interrupt.
							; 33617 - on / off

	defw	0x0038			; IM1 routine address.
	defb	1, 7, 1			; activated, trigger every interrupt.
							; 33622 - on / off

	defs	50, 0			; Leave rest of table empty.

; ------------------------------------------------------------------------------
; RAINBOW PROCESSOR
; Uses LD / PUSH to set 18 columns of hi-colour per scanline.

rainbow_proc:
	ld		bc, 0x7ffd		; Paging.
	ld		a, %00010000	; RAM 0 | SCR 0 | ROM 1
	out		(c), a
	ld		(sp_save+1), sp	; Save stack pointer.
	ld		b, 7
	call	drawing			;
	ld		b, 54			;

delay:
	djnz	delay			;
	add		hl, hl			;
	nop						; We should be close to cycle 15900 at this point.

rainbow:					; This repeated block expands to 5094 bytes!
REPT 144, line
	ld		sp, 0x5820+((line*4) & ~0x1f)+grid_offset+(num_cols*2)
	ld		hl, 0x0000			;0x0800
	ld		de, 0x0000			;0x1810
	ld		bc, 0x0000			;0x2820
	exx
	ld		hl, 0x0000			;0x3028
	ld		de, 0x0000			;0x0800
	ld		bc, 0x0000			;0x1810
	push	bc
	push	de
	push	hl
	ld		hl, 0x0000			;0x0030
	ld		de, 0x0000			;0x1008
	ld		bc, 0x0000			;0x2018
	push	bc
	push	de
	push	hl
	exx
	push	bc
	push	de
	push	hl
ENDM
sp_save:
	ld		sp, 0
	ld		b, 2				; draw another two tiles for a total of 9 per interrupt

drawing:
	ld		a, (tile_num)

draw_lp:
	push	bc
	call	draw_tile
	ld		hl, tile_num
	ld		a, (hl)
	inc		a
	ld		c, a
	cp		num_tiles
	sbc		a, a
	and		c
	ld		(hl), a
	pop		bc
	djnz	draw_lp
	ret

; ------------------------------------------------------------------------------
; DRAW TILE
; Draw a single tile on the screen.

draw_tile:						; draw tile in A (*timing sensitive*!)
	ld		c, a				; save tile
	ld		l, a
	ld		h, tile_map/256		; tile map entry for tile number
	ld		a, (hl)				; tile value
	ld		l, 0
	srl		a
	rr		l					; /2
	rra
	rr		l					; /4
	ld		h, a				; *64
	ld		de, tile_gfx
	add		hl, de
	ld		a, c				; restore tile
	exx
	add		a, a				; *2
	add		a, rowcol_tab%256
	ld		l, a
	ld		h, rowcol_tab/256	; tile to row/col table
	ld		a, (hl)				; row *2
	add		a, a				; *4
	add		a, (hl)				; *6
	add		a, addr_tab%256
	ld		e, a
	ld		d, addr_tab/256	; row in address look-up table
	inc		l
	ld		a, (hl)				; col
	ld		c, a				; save col
	add		a, offset_tab%256
	ld		l, a
	ld		h, offset_tab/256	; column in offset table
	ld		a, c				; restore col
	ld		c, (hl)				; column offset in rainbow code
	ld		b, 0
	ex		de, hl
	ld		e, (hl)
	inc		l
	ld		d, (hl)				; attr base for row
	inc		l
	ex		de, hl
	add		hl, bc				; tile attr addr
	push	hl					; keep for drawing below
	ex		de, hl
	add		a, a				; col *2
	ld		c, a				; bc=data offset in row
REPT 2
	ld		e, (hl)
	inc		l
	ld		d, (hl)				; display base for row
	inc		l
	ex		de, hl
	add		hl, bc				; tile display addr(s)
	push	hl
	ex		de, hl
ENDM
	exx
	ld		bc, 32

data_lp:
	pop	de						; next tile display address
REPT 7
	ldi							; copy 2 bytes
	ldi
	dec	e
	dec	e
	inc	d						; next line
ENDM
	ldi							; final bytes
	ldi
	jp		pe, data_lp
	pop		de					; tile attr address
	inc	b						; ensure B zero after first LDI
REPT 15
	ldi							; copy 2 bytes
	ldi
	ex		de, hl
	ld		c, 41-2				; offset to next rainbow code line
	add		hl, bc
	ex		de, hl
ENDM
	ldi							; final attrs
	ldi
	ret

; ------------------------------------------------------------------------------
; ANIMATION
; Called on every 18th interrupt

scroll_tiles:
	ld		hl, (tile_gfx+30)	; water
	exx
	ld		hl, tile_gfx+29
	ld		de, tile_gfx+31
	call	scroll_30
	exx
	ld		(tile_gfx), hl

	ld		hl, (tile_gfx+30+64)
	exx
	ld		hl, tile_gfx+29+64
	ld		de, tile_gfx+31+64
	call	scroll_30
	exx
	ld		(tile_gfx+64), hl

	ld		hl, (tile_gfx+30+128)
	exx
	ld		hl, tile_gfx+29+128
	ld		de, tile_gfx+31+128
	call	scroll_30
	exx
	ld		(tile_gfx+128), hl

	ld		hl, (tile_gfx+30+4352)	; magic fields
	exx
	ld		hl, tile_gfx+29+4352
	ld		de, tile_gfx+31+4352
	call	scroll_30
	exx
	ld		(tile_gfx+4352), hl

	ld		hl, (tile_gfx+30+4416)
	exx
	ld		hl, tile_gfx+29+4416
	ld		de, tile_gfx+31+4416
	call	scroll_30
	exx
	ld		(tile_gfx+4416), hl

	ld		hl, (tile_gfx+30+4480)
	exx
	ld		hl, tile_gfx+29+4480
	ld		de, tile_gfx+31+4480
	call	scroll_30
	exx
	ld		(tile_gfx+4480), hl

	ld		hl, (tile_gfx+30+4544)
	exx
	ld		hl, tile_gfx+29+4544
	ld		de, tile_gfx+31+4544
	call	scroll_30
	exx
	ld		(tile_gfx+4544), hl

	ret

animate_flags:
	ld		hl, tile_gfx+646	; town
	ld		de, tile_gfx+648
	ld		a, (de)
	ld		b, (hl)
	ld		(hl), a
	ld		a, b
	ld		(de), a

	ld		hl, tile_gfx+707	; castle
	ld		de, tile_gfx+709
	ld		a, (de)
	ld		b, (hl)
	ld		(hl), a
	ld		a, b
	ld		(de), a

	ld		hl, tile_gfx+898	; LB castle
	ld		de, tile_gfx+900
	ld		a, (de)
	ld		b, (hl)
	ld		(hl), a
	ld		a, b
	ld		(de), a

	ld		hl, tile_gfx+1028	; player ship west
	ld		de, tile_gfx+1030
	ld		a, (de)
	ld		b, (hl)
	ld		(hl), a
	ld		a, b
	ld		(de), a

	ld		hl, tile_gfx+1029
	ld		de, tile_gfx+1031
	ld		a, (de)
	ld		b, (hl)
	ld		(hl), a
	ld		a, b
	ld		(de), a

	ld		hl, tile_gfx+1061
	ld		de, tile_gfx+1063
	ld		a, (de)
	ld		b, (hl)
	ld		(hl), a
	ld		a, b
	ld		(de), a

	ld		hl, tile_gfx+1156	; player ship east
	ld		de, tile_gfx+1158
	ld		a, (de)
	ld		b, (hl)
	ld		(hl), a
	ld		a, b
	ld		(de), a
	
	ld		hl, tile_gfx+1188
	ld		de, tile_gfx+1190
	ld		a, (de)
	ld		b, (hl)
	ld		(hl), a
	ld		a, b
	ld		(de), a
	
	ld		hl, tile_gfx+1157
	ld		de, tile_gfx+1159
	ld		a, (de)
	ld		b, (hl)
	ld		(hl), a
	ld		a, b
	ld		(de), a
	ret

animate_campfire:
	ld		a, (tile_gfx+4784+64)
	ld		c, a
	ld		a, (tile_gfx+4786+64)
	ld		(tile_gfx+4784+64), a
	ld		(tile_gfx+4785+64), a
	ld		(tile_gfx+4788+64), a
	ld		(tile_gfx+4789+64), a
	ld		(tile_gfx+4792+64), a
	ld		(tile_gfx+4793+64), a
	ld		a, c
	ld		(tile_gfx+4786+64), a
	ld		(tile_gfx+4787+64), a
	ld		(tile_gfx+4790+64), a
	ld		(tile_gfx+4791+64), a

	ret

swap_tiles:
	ld		hl, tile_map
	ld		b, 81

anim_lp:
	ld		a, (hl)
	ld		c, a
	and		%11110000
	cp		%00100000			; 32 to 47?
	jr		z, anim_2
	cp		%01010000			; 80 to 95?
	jr		z, anim_2
	ld		a, c
	sub		132
	and		%11110000			; 132 to 143?
	jr		z, anim_2
	ld		a, c
	cp		144
	jr		c, done_anim
	inc		c					; next frame
	ld		a, c
	and		%00000011			; looped?
	ld		a, c
	jr		nz, done_anim
	sub		4					; loop back
	jr		done_anim

anim_2:
	ld		a, c
	xor		1

done_anim:
	ld		(hl), a
	inc		l
	djnz	anim_lp
	ret

scroll_30:
REPT 30
	ldd
ENDM
	ret

; ------------------------------------------------------------------------------
; PROTRACKER 3 EXTREME II (P3X) AY PLAYER
; For ZX81 (Zon-X), TC2068, and Spectrum 128.

; Copyright (c) 2004, 2007 S.V.Bulba <vorobey@mail.khstu.ru>

; http://bulba.untergrund.net (http://bulba.at.kz)
; Formatted by Gasman for Pasmo and optimized by Cheveron

; variables are 541 bytes long and follow the player
; music is in ProTracker3.6 format (minus the 100 byte header)

checklp:
	ld	hl, setup
	set	7, (hl)
	bit	0, (hl)
	ret	z
	pop	hl
	ld	hl, delycnt
	inc	(hl)
	ld	hl, chana + chp_ntskcn
	inc	(hl)

mute:
	xor	a
	ld	h, a
	ld	l, a	 
	ld	(ayregs + ar_ampla), a
	ld	(ayregs + ar_amplb), hl
	jp	rout_a0

p3x_init:
	ld hl, music - 100	; Music address.
	ld	(modaddr + 1), hl
	ld	(mdaddr2 + 1), hl
	push	hl
	ld	de, 100
	add	hl, de
	ld	a, (hl)
	ld	(p3x_delay + 1), a
	push	hl
	pop	ix
	add	hl, de
	ld	(crpsptr), hl
	ld	e, (ix + 2)
	add	hl, de
	inc	hl
	ld	(lposptr + 1), hl
	pop	de
	ld	l, (ix + 3)
	ld	h, (ix + 4)
	add	hl, de
	ld	(patsptr + 1), hl
	ld	hl, 169
	add	hl, de
	ld	(ornptrs + 1), hl
	ld	hl, 105
	add	hl, de
	ld	(samptrs + 1), hl
	ld	hl, setup
	res	7, (hl)
	xor	a
	ld	hl, vars
	ld	(hl), a
	ld	de, vars + 1
	ld	bc, var0end - vars - 1
	ldir
	inc	a
	ld	(delycnt), a
	ld	hl, 0f001h
	ld	(chana + chp_ntskcn), hl
	ld	(chanb + chp_ntskcn), hl
	ld	(chanc + chp_ntskcn), hl
	ld	hl, emptysamorn
	ld	(adinpta + 1), hl
	ld	(chana + chp_ornptr), hl
	ld	(chanb + chp_ornptr), hl
	ld	(chanc + chp_ornptr), hl
	jp	rout_a0

pd_orsm:
	ld	(ix + 8), 0		; pattern decoder
	call	setorn
	ld	a, (bc)
	inc	bc
	rrca

pd_sam:
	add	a, a

pd_sam_:
	ld	e, a
	ld	d, 0

samptrs:
	ld	hl, 8481
	add	hl, de
	ld	e, (hl)
	inc	hl
	ld	d, (hl)

modaddr
	ld	hl, 8481
	add	hl, de
	ld	(ix + 3), l
	ld	(ix + 4), h
	jr	pd_loop

pd_vol:
	rlca
	rlca
	rlca
	rlca
	ld	(ix + 16), a
	jr	pd_lp2

pd_eoff:
	ld	(ix + 8), a
	ld	(ix - 12), a
	jr	pd_lp2

pd_sore:
	dec	a
	jr	nz, pd_env
	ld	a, (bc)
	inc	bc
	ld	(ix + 5), a
	jr	pd_lp2

pd_env:
	call	setenv
	jr	pd_lp2

pd_orn:
	call	setorn
	jr	pd_loop

pd_esam:
	ld	(ix + 8), a
	ld	(ix - 12), a
	call	nz, setenv
	ld	a, (bc)
	inc	bc
	jr	pd_sam_

ptdecod:
	ld	a, (ix + 6)
	ld	(prnote + 1), a
	ld	l, (ix - 6)
	ld	h, (ix - 5)
	ld	(prslide + 1), hl

pd_loop:
	ld	de, 8208

pd_lp2:
	ld	a, (bc)
	inc	bc
	add	a, e
	jr	c, pd_orsm
	add	a, d
	jr	z, pd_fin
	jr	c, pd_sam
	add	a, e
	jr	z, pd_rel
	jr	c, pd_vol
	add	a, e
	jr	z, pd_eoff
	jr	c, pd_sore
	add	a, 060h
	jr	c, pd_note
	add	a, e
	jr	c, pd_orn
	add	a, d
	jr	c, pd_nois
	add	a, e
	jr	c, pd_esam
	add	a, a
	ld	e, a
	ld	hl, spccoms + 57120
	add	hl, de
	ld	e, (hl)
	inc	hl
	ld	d, (hl)
	push	de
	jr	pd_loop

pd_nois:
	ld	(ns_base), a
	jr	pd_lp2

pd_rel:
	res	0, (ix + 9)
	jr	pd_res

pd_note:
	ld	(ix + 6), a
	set	0, (ix + 9)
	xor	a

pd_res:
	ld	(pdsp_ + 1), sp
	ld	sp, ix
	ld	h, a
	ld	l, a
	push	hl
	push	hl
	push	hl
	push	hl
	push	hl
	push	hl

pdsp_
	ld	sp, 12593

pd_fin:
	ld	a, (ix + 5)
	ld	(ix + 15), a
	ret

c_portm:
	res	2, (ix + 9)
	ld	a, (bc)
	inc	bc
	inc	bc
	inc	bc
	ld	(ix + 10), a
	ld	(ix - 7), a
	ld	de, nt_
	ld	a, (ix + 6)
	ld	(ix + 7), a
	add	a, a
	ld	l, a
	ld	h, 0
	add	hl, de
	ld	a, (hl)
	inc	hl
	ld	h, (hl)
	ld	l, a
	push	hl

prnote:
	ld	a, 62
	ld	(ix + 6), a
	add	a, a
	ld	l, a
	ld	h, 0
	add	hl, de
	ld	e, (hl)
	inc	hl
	ld	d, (hl)
	pop	hl
	sbc	hl, de
	ld	(ix + 13), l
	ld	(ix + 14), h
	ld	e, (ix - 6)
	ld	d, (ix - 5)

prslide:
	ld	de, 4369
	ld	(ix - 6), e
	ld	(ix - 5), d
	ld	a, (bc)
	inc	bc
	ex	af, af'	;'
	ld	a, (bc)
	inc	bc
	and	a
	jr	z, nosig
	ex	de, hl

nosig:
	sbc	hl, de
	jp	p, set_stp
	cpl
	ex	af, af'	;'
	neg
	ex	af, af'	;'

set_stp:
	ld	(ix + 12), a
	ex	af, af'	;'
	ld	(ix + 11), a
	ld	(ix - 2), 0
	ret

c_gliss:
	set	2, (ix + 9)
	ld	a, (bc)
	inc	bc
	ld	(ix + 10), a
	and	a
	jr	nz, gl36
	ld	a, 6			; P3X version is always PT3.6
	cp	7
	sbc	a, a
	inc	a

gl36:
	ld	(ix - 7), a
	ld	a, (bc)
	inc	bc
	ex	af, af'	;'
	ld	a, (bc)
	inc	bc
	jr	set_stp

c_smpos:
	ld	a, (bc)
	inc	bc
	ld	(ix - 11), a
	ret

c_orpos:
	ld	a, (bc)
	inc	bc
	ld	(ix - 12), a
	ret

c_vibrt:
	ld	a, (bc)
	inc	bc
	ld	(ix - 1), a
	ld	(ix - 2), a
	ld	a, (bc)
	inc	bc
	ld	(ix + 0), a
	xor a
	ld	(ix - 7), a
	ld	(ix - 6), a
	ld	(ix - 5), a
	ret

c_engls
	ld	a, (bc)
	inc	bc
	ld	(env_del + 1), a
	ld	(curedel), a
	ld	a, (bc)
	inc	bc
	ld	l, a
	ld	a, (bc)
	inc	bc
	ld	h, a
	ld	(esldadd + 1), hl
	ret

c_p3x_delay:
	ld	a, (bc)
	inc	bc
	ld	(p3x_delay + 1), a
	ld	(delycnt), a		; bugfix by Lee du Caine
	ret

setenv:
	ld	(ix + 8), e
	ld	(ayregs + ar_envtp), a
	ld	a, (bc)
	inc	bc
	ld	h, a
	ld	a, (bc)
	inc	bc
	ld	l, a
	ld	(envbase), hl
	xor	a
	ld	(ix - 12), a
	ld	(curedel), a
	ld	h, a
	ld	l, a
	ld	(curesld), hl
c_nop:
	ret

setorn:
	add	a, a
	ld	e, a
	ld	d, 0
	ld	(ix - 12), d

ornptrs:
	ld	hl, 8481
	add	hl, de
	ld	e, (hl)
	inc	hl
	ld	d, (hl)

mdaddr2:
	ld	hl, 8481
	add	hl, de
	ld	(ix + 1), l
	ld	(ix + 2), h
	ret

;all 16 addresses to protect from broken pt3 modules

spccoms:
	defw	c_nop
	defw	c_gliss
	defw	c_portm
	defw	c_smpos
	defw	c_orpos
	defw	c_vibrt
	defw	c_nop
	defw	c_nop
	defw	c_engls
	defw	c_p3x_delay
	defw	c_nop
	defw	c_nop
	defw	c_nop
	defw	c_nop
	defw	c_nop
	defw	c_nop

chregs:
	xor	a
	ld	(ampl), a
	bit	0, (ix + 21)
	push	hl
	jp	z, ch_exit
	ld	(csp_ + 1), sp
	ld	l, (ix + 13)
	ld	h, (ix + 14)
	ld	sp, hl
	pop	de
	ld	h, a
	ld	a, (ix + 0)
	ld	l, a
	add	hl, sp
	inc	a
	cp	d
	jr	c, ch_orps
	ld	a, e

ch_orps:
	ld	(ix + 0), a
	ld	a, (ix + 18)
	add	a, (hl)
	jp	p, ch_ntp
	xor a

ch_ntp:
	cp	96
	jr	c, ch_nok
	ld	a, 95

ch_nok:
	add	a, a
	ex	af, af'	;'
	ld	l, (ix + 15)
	ld	h, (ix + 16)
	ld	sp, hl
	pop	de
	ld	h, 0
	ld	a, (ix + 1)
	ld	b, a
	add	a, a
	add	a, a
	ld	l, a
	add	hl, sp
	ld	sp, hl
	ld	a, b
	inc	a
	cp	d
	jr	c, ch_smps
	ld	a, e

ch_smps:
	ld	(ix + 1), a
	pop	bc
	pop	hl
	ld	e, (ix + 8)
	ld	d, (ix + 9)
	add	hl, de
	bit	6, b
	jr	z, ch_noac
	ld	(ix + 8), l
	ld	(ix + 9), h

ch_noac:
	ex	de, hl
	ex	af, af'	;'
	ld	l, a
	ld	h, 0
	ld	sp, nt_
	add	hl, sp
	ld	sp, hl
	pop	hl
	add	hl, de
	ld	e, (ix + 6)
	ld	d, (ix + 7)
	add	hl, de

csp_:
	ld	sp, 03131h
	ex	(sp), hl
	xor	a
	or	(ix + 5)
	jr	z, ch_amp
	dec	(ix + 5)
	jr	nz, ch_amp
	ld	a, (ix + 22)
	ld	(ix + 5), a
	ld	l, (ix + 23)
	ld	h, (ix + 24)
	ld	a, h
	add	hl, de
	ld	(ix + 6), l
	ld	(ix + 7), h
	bit	2, (ix + 21)
	jr	nz, ch_amp
	ld	e, (ix + 25)
	ld	d, (ix + 26)
	and	a
	jr	z, ch_stpp
	ex	de, hl

ch_stpp:
	sbc	hl, de
	jp	m, ch_amp
	ld	a, (ix + 19)
	ld	(ix + 18), a
	xor	a
	ld	(ix + 5), a
	ld	(ix + 6), a
	ld	(ix + 7), a

ch_amp:
	ld	a, (ix + 2)
	bit	7, c
	jr	z, ch_noam
	bit	6, c
	jr	z, ch_amin
	cp	15
	jr	z, ch_noam
	inc	a
	jr	ch_svam

ch_amin:
	cp	241
	jr	z, ch_noam
	dec	a

ch_svam:
	ld	(ix + 2), a

ch_noam:
	ld	l, a
	ld	a, b
	and	15
	add	a, l
	jp	p, ch_apos
	xor	a

ch_apos:
	cp	16
	jr	c, ch_vol
	ld	a, 15

ch_vol:
	or	(ix + 28)
	ld	l, a
	ld	h, 0
	ld	de, vt_
	add	hl, de
	ld	a, (hl)

ch_env:
	bit	0, c
	jr	nz, ch_noen
	or	(ix + 20)

ch_noen:
	ld	(ampl), a
	bit	7, b
	ld	a, c
	jr	z, no_ensl
	rla
	rla
	sra	a
	sra	a
	sra	a
	add	a, (ix + 4)
	bit	5, b
	jr	z, no_enac
	ld	(ix + 004h), a

no_enac:
	ld	hl, addtoen + 1
	add	a, (hl)
	ld	(hl), a
	jr	ch_mix

no_ensl:
	rra
	add	a, (ix + 3)
	ld	(addtons), a
	bit	5, b
	jr	z, ch_mix
	ld	(ix + 3), a

ch_mix:
	ld	a, b
	rra
	and	72

ch_exit:
	ld	hl, ayregs + ar_mixer
	or	(hl)
	rrca
	ld	(hl), a
	pop	hl
	xor	a
	or	(ix + 10)
	ret	z
	dec	(ix + 10)
	ret	nz
	xor	(ix + 21)
	ld	(ix + 21), a
	rra
	ld	a, (ix + 11)
	jr	c, ch_ondl
	ld	a, (ix + 12)

ch_ondl:
	ld	(ix + 10), a
	ret

play:
	xor	a
	ld	(addtoen + 1), a
	ld	(ayregs + ar_mixer), a
	dec	a
	ld	(ayregs + ar_envtp), a
	ld	hl, delycnt
	dec	(hl)
	jr	nz, pl2
	ld	hl, chana + chp_ntskcn
	dec	(hl)
	jr	nz, pl1b

adinpta:
	ld	bc, 00101h
	ld	a, (bc)
	and	a
	jr	nz, pl1a
	ld	d, a
	ld	(ns_base), a
crpsptr equ $+1
	ld	hl, 0
	inc	hl
	ld	a, (hl)
	inc	a
	jr	nz, plnlp
	call	checklp

lposptr:
	ld	hl, 8481
	ld	a, (hl)
	inc	a
plnlp:
	ld	(crpsptr), hl
	dec	a
	add	a, a
	ld	e, a
	rl	d

patsptr:
	ld	hl, 8481
	add	hl, de
	ld	de, (modaddr + 1)
	ld	(psp_ + 1), sp
	ld	sp, hl
	pop	hl
	add	hl, de
	ld	b, h
	ld	c, l
	pop	hl
	add	hl, de
	ld	(adinptb + 1), hl
	pop	hl
	add	hl, de
	ld	(adinptc + 1), hl

psp_:
	ld	sp, 12593

pl1a:
	ld	ix, chana + 12
	call	ptdecod
	ld	(adinpta + 1), bc

pl1b:
	ld	hl, chanb + chp_ntskcn
	dec	(hl)
	jr	nz, pl1c
	ld	ix, chanb + 12

adinptb:
	ld	bc, 257
	call	ptdecod
	ld	(adinptb + 1), bc

pl1c:
	ld	hl, chanc + chp_ntskcn
	dec	(hl)
	jr	nz, p3x_delay
	ld	ix, chanc + 12

adinptc:
	ld	bc, 257
	call	ptdecod
	ld	(adinptc + 1), bc

p3x_delay:
	ld	a, 62
	ld	(delycnt), a

pl2:
	ld	ix, chana
	ld	hl, (ayregs + ar_tona)
	call	chregs
	ld	(ayregs + ar_tona), hl
	ld	a, (ampl)
	ld	(ayregs + ar_ampla), a
	ld	ix, chanb
	ld	hl, (ayregs + ar_tonb)
	call	chregs
	ld	(ayregs + ar_tonb), hl
	ld	a, (ampl)
	ld	(ayregs + ar_amplb), a
	ld	ix, chanc
	ld	hl, (ayregs + ar_tonc)
	call	chregs
	ld	(ayregs + ar_tonc), hl
	ld	hl, (ns_base_addtons)
	ld	a, h
	add	a, l
	ld	(ayregs + ar_noise), a

addtoen:
	ld	a, 62
	ld	e, a
	add	a, a
	sbc	a, a
	ld	d, a
	ld	hl, (envbase)
	add	hl, de
	ld	de, (curesld)
	add	hl, de
	ld	(ayregs + ar_env), hl
	xor	a
	ld	hl, curedel
	or	(hl)
	jr	z, rout_a0
	dec	(hl)
	jr	nz, rout

env_del:
	ld	a, 62
	ld	(hl), a

esldadd:
	ld	hl, 8481
	add	hl, de
	ld	(curesld), hl

rout:
	xor	a

rout_a0:
;	ld	de, #df0f		; Zon-X ports
;	ld	bc, #ffdf		;

;	ld	de, #f5f6		; TC2068 ports
;	ld	bc, #fff5		;

	ld	de, #ffbf		; 128 ports
	ld	bc, #fffd		;

	ld	hl, ayregs

lout:
	out	(c), a
;	ld	c, e			; Zon-X / TC2068
	ld	b, e			; 128
	outi
;	ld	c, d			; Zon-X / TC2068
	ld	b, d			; 128
	inc	a
	cp	00dh
	jr	nz, lout
	out	(c), a
	ld	a, (hl)
	and	a
	ret	m
;	ld	c, e			; Zon-X / TC2068
	ld	b, e			; 128
	out	(c), a
	ret

t_:

tcold_0:
	defb	#00 + 1, #04 + 1, #08 + 1, #0a + 1, #0c + 1, #0e + 1
	defb	#12 + 1, #14 + 1, #18 + 1, #24 + 1, #3c + 1, 0
tcold_1:
	defb	#5c + 1, 0

tcold_2:
	defb	#30 + 1, #36 + 1, #4c + 1, #52 + 1, #5e + 1, #70 + 1
	defb	#82, #8c, #9c, #9e, #a0, #a6, #a8, #aa, #ac, #ae, #ae, 0

tcnew_3:
	defb	#56 + 1

tcold_3:
	defb	#1e + 1, #22 + 1, #24 + 1, #28 + 1, #2c + 1, #2e + 1
	defb	#32 + 1, #be + 1, 0

tcnew_0:
	defb	#1c + 1, #20 + 1, #22 + 1, #26 + 1, #2a + 1, #2c + 1
	defb	#30 + 1, #54 + 1, #bc + 1, #be + 1, 0

tcnew_1 equ tcold_1

tcnew_2:
	defb	#1a + 1, #20 + 1, #24 + 1, #28 + 1, #2a + 1, #3a + 1
	defb	#4c + 1, #5e + 1, #ba + 1, #bc + 1, #be + 1, 0

emptysamorn equ $ - 1

	defb	1, 0;, #90 ;delete #90 if you don't need default sample

; channel data offsets

chp_psinor	equ 0
chp_psinsm	equ 1
chp_cramsl	equ 2
chp_crnssl	equ 3
chp_crensl	equ 4
chp_tslcnt	equ 5
chp_crtnsl	equ 6
chp_tnacc	equ 8
chp_conoff	equ 10
chp_onoffd	equ 11			; ix for ptdecod here ( + 12)
chp_offond	equ 12
chp_ornptr	equ 13
chp_samptr	equ 15
chp_nntskp	equ 17
chp_note	equ 18
chp_sltont	equ 19
chp_env_en	equ 20
chp_flags	equ 21			; enabled - 0, simplegliss - 2
chp_tnsldl	equ 22
chp_tslstp	equ 23
chp_tndelt	equ 25
chp_ntskcn	equ 27
chp_volume	equ 28

vars:
chana	ds 29
chanb	ds 29
chanc	ds 29
delycnt	defb	0			; globalvars
curesld	defw	0
curedel	defb	0

ns_base_addtons:
ns_base	defb	0
addtons	defb	0

ar_tona equ 0				; word 1
ar_tonb equ 2				; word 1
ar_tonc equ 4				; word 1
ar_noise equ 6				; byte 1
ar_mixer equ 7				; byte 1
ar_ampla equ 8				; byte 1
ar_amplb equ 9				; byte 1
ar_amplc equ 10				; byte 1
ar_env equ 11				; word 1
ar_envtp equ 13				; byte 1
ar_size equ 14				; byte 1

ayregs equ $

vt_:
	incbin	"volume.p3v"

envbase equ vt_ + 14

t1_ equ vt_ + 16			; tone tables data depacked here

t_old_1	equ t1_
t_old_2	equ t_old_1 + 24
t_old_3	equ t_old_2 + 24
t_old_0	equ t_old_3 + 2
t_new_0	equ t_old_0
t_new_1	equ t_old_1
t_new_2	equ t_new_0 + 24
t_new_3	equ t_old_3

nt_:
	incbin	"notes.p3n"

;local vars
ampl equ ayregs + ar_amplc
var0end	equ vt_ + 16			; p3x_init zeroes from vars to
					; var0end - 1
varsend equ $
mdladdr equ $

; ------------------------------------------------------------------------------
; CHARACTER DISPLAY DRIVER
; Use PRINT #4

scrollw:
	ld		ix, row_table	; scroll text window up one row
	ld		bc, 96*256+12*96
	di						; temporarily disable interrupts
	
scroll_loop:
	ld		e, (ix + 0)
	ld		d, (ix + 1)
	ld		l, (ix + 16)
	ld		h, (ix + 17)
	ldi
	ldi
	ldi
	ldi
	ldi
	ldi
	ldi
	ldi
	ldi
	ldi
	ldi
	ldi
	inc		ix
	inc		ix
	djnz	scroll_loop
	ei						; restore interrupts

dwncr:
	ld		hl, 160			; back to the bottom left hand side of text window.
	ld		(coords_x), hl
	ret

print6x8:
	push	hl				; Save all the	registers, call the 
	push	bc				; printing routine, put all the registers
	push	de				; back again and leap back to the
	push	af				; operating system.
	call	doit
	pop		af
	pop		de
	pop		bc
	pop		hl
	ret

doit:
	ld		hl, atflg		; Look to see if last character was a
	dec		(hl)			; control code 22 - the code for AT.
	jp		m, atchq
	jr		z, getxp

getyp:
	sub		185				; A - 185
	cpl						; 184 - A
	ld		(coords_y), a
	ret

getxp:
	ld		(coords_x), a	; If the last charater was an AT then fetch
	ret						; the X and Y co-ordinates and move to the
							; new printing position.

atchq:
	cp		22				; If the current character is an AT control
	jr		nz, crchq		; code, set the ATFLG to indicate that the
	ld		(hl), 2			; next two codes dealt with will be the X
	ret						; and Y positions for the print.

crchq:
	inc		(hl)
	cp		6				; If the current character is a tab
	jr		z, scrollw		; control code, scroll the text window.
	
	cp		13				; If the current character is a return 
	jr		z, dwncr		; control code, reset the position.
	
get_addr:
	ld		h, 0			; Figure out where the character's
	ld		l, a			; definition is stored in memory.
	add		hl, hl
	add		hl, hl
	add		hl, hl
	ex		de, hl
	ld		hl, font - 256 + 7

	add		hl, de			; the address of the seventh byte of the character

	ld		(chrad), hl		; calculate the address
	ld		a, (coords_y)	; in the display file where the character
	cp		185				; will be printed.
	jr		c, noterr5
	rst		8
	defb	4

noterr5:
	ld		bc, (coords_x)
	ld		a, 191			; use entire display
	call	pixel_add + 2
	ld		(s_posn), a
	ld		(df_cc), hl
	ld		b, 8

prnlp:
	push	bc
	ld		hl, (chrad)
	ld		a, (hl)
	dec		hl
	ld		(chrad), hl
	ld		l, a
	ld		a, (s_posn)
	ld		bc, 0xff03		; mask
	push	bc				; save mask
	and		a				; sometimes no rotation is required
	jr		z, putit
	pop		bc				; drop saved mask
	ld		b, a
	push	bc				; save count	
	ex		de, hl			; save character

mask:
	ld		hl, 0x00fc

rotmsk:
	srl		l
	rr		h
	djnz	rotmsk
	ld		a, 255			; invert mask
	xor		h
	ld		h, a
	ld		a, 255
	xor		l
	ld		l, a
	pop		bc				; restore count
	push	hl				; save mask
	ex		de, hl			; restore character
	ld		h, 0

rotlp:
	srl		l				; Rotate the character definition into the 
	rr		h				; correct pixel position, place it on the
	djnz	rotlp			; screen one byte at a time.

putit:
	pop		bc				; restore mask
	ld		de, (df_cc)
	ld		a, (de)
	and		c				; mask off
	xor		l
	ld		(de), a
	ld		a, (s_posn)
	and		a
	jr		z, pst
	inc		e
	ld		a, e
	and		31
	jr		z, pst
	ld		a, (coords_x)
	or		248
	add		a, 4			; always 6 pixels 
	jr		nc, pst
	ld		a, (de)
	and		b				; mask off
	xor		h
	ld		(de), a

pst:
	ld		hl, (df_cc)
	call	uline
	ld		(df_cc), hl
	pop		bc
	djnz	prnlp
	ld		hl, coords_x
	ld		a, (hl)			; the character just printed.
	add		a, 6			; always 6 pixels
	ld		(hl), a
	ret

uline:
	push	af				; Here's a handy routine that sets HL to
	ld		a, h			; point to the next pixel line up in the
	dec		h				; display file.
	and		7
	jr		nz, _end
	ld		a, l
	sub		32
	ld		l, a
	jr		c, _end
	ld		a, h
	add		a, 8
	ld		h, a

_end:
	pop		af
	ret

; ------------------------------------------------------------------------------
; PLAY SOUND FX
; POKE the value from BASIC or call playfx with the value in BC

playBasic:
	ld	bc, 0

playfx:
	ld hl,soundEffectsData	;address of sound effects data

	di
	push iy

	add hl,bc
	add hl,bc
	ld e,(hl)
	inc hl
	ld d,(hl)
	push de
	pop ix			;put it into ix

readData:
	ld a,(ix)		;read block type
	or a
	jr nz,_sound
	pop iy
	ei
	ret

_sound:
	ld c,(ix+1)		;read duration 1
	ld b,(ix+2)
	ld e,(ix+3)		;read duration 2
	ld d,(ix+4)
	push de
	pop iy

	dec a
	jr nz,sfxRoutineNoise

;this routine generate tone with many parameters

sfxRoutineTone:
	ld e,(ix+5)		;freq
	ld d,(ix+6)
	ld a,(ix+9)		;duty
	ld (_duty),a
	ld hl,0

_l0:
	push bc
	push iy
	pop bc
_l1:
	add hl,de
	ld a,h
_duty equ $+1
	cp 0
	sbc a,a
	and 16
_border equ $+1
;	or 0
	out (#fe),a

	dec bc
	ld a,b
	or c
	jr nz,_l1

	ld a,(_duty)
	add a,(ix+10)	;duty change
	ld (_duty),a

	ld c,(ix+7)		;slide
	ld b,(ix+8)
	ex de,hl
	add hl,bc
	ex de,hl

	pop bc
	dec bc
	ld a,b
	or c
	jr nz,_l0

	ld c,11

nextData:
	add ix,bc		;skip to the next block
	jr readData

;this routine generate noise with two parameters

sfxRoutineNoise:
	ld e,(ix+5)		;pitch

	ld d,1
	ld h,d
	ld l,d

_l0:
	push bc
	push iy
	pop bc

_l1:
	ld a,(hl)
	and 16
_border equ $+1
;	or 0
	out (#fe),a
	dec d
	jr nz,_l2
	ld d,e
	inc hl
	ld a,h
	and #1f
	ld h,a
_l2:
	dec bc
	ld a,b
	or c
	jr nz,_l1

	ld a,e
	add a,(ix+6)	;slide
	ld e,a

	pop bc
	dec bc
	ld a,b
	or c
	jr nz,_l0

	ld c,7
	jr nextData

; ------------------------------------------------------------------------------
; SOUND FX DATA

soundEffectsData:

	org		0xbfff
	defb	0
	
; ------------------------------------------------------------------------------
; BANK 7
